@@ -1,8 +1,9 @@
 #!/usr/bin/env python2.7
 # coin.py: a coin implemented on top of pybc
 
-import struct, hashlib, traceback, shelve, time
+import struct, hashlib, traceback, time, threading
 import pybc, pyelliptic
+import sqliteshelf
 
 class Transaction(object):
     """
@@ -574,6 +575,14 @@ class CoinBlockchain(pybc.TransactionalBlockchain):
         # CoinState to track unspent outputs.
         super(CoinBlockchain, self).__init__(pybc.PowAlgorithm(), block_store, 
             initial_state=CoinState())
+            
+        # Set up the blockchain for 1 minute blocks, retargeting every 10
+        # blocks
+        # This is in blocks
+        self.retarget_period = 10
+        # This is in seconds
+        self.retarget_time = self.retarget_period * 60
+        
     
     def transaction_valid_for_relay(self, transaction_bytes):
         """
@@ -597,15 +606,27 @@ class CoinBlockchain(pybc.TransactionalBlockchain):
         
         """
         
-        # Easy example: constant block reward.
-        return 50
+        # Easy example: 50 coins forever
+        
+        # Get the height of this block
+        if previous_block is not None:
+            height = previous_block.height + 1
+        else:
+            height = 0
+
+        # How many coins should we generate? We could do something based on
+        # height, but just be easy.
+        coins = 50
+        
+        # Return the number of coins to generate
+        return coins
     
     def verify_transaction(self, transaction_bytes, chain_head, 
         other_transactions=[]):
         """
-        If the given Transaction is valid atop the given Block chain_head and in
-        light of the already verified other transactions, return True.
-        Otherwise, return False.
+        If the given Transaction is valid atop the given Block chain_head (which
+        may be None) and in light of the already verified other transactions,
+        return True. Otherwise, return False.
         
         Ensures that:
         
@@ -755,12 +776,17 @@ class Wallet(object):
         
         """
         
-        # Use a shelve shelf to keep pyelliptic ECC objects for our addresses.
-        # keypairs are stored by public key hash.
-        self.keystore = shelve.open(filename)
+        # Use a pdict database to keep pyelliptic ECC objects for our addresses.
+        # keypairs are stored by public key hash. Make it not lazy so it saves
+        # constantly, so we don't lose keys.
+        self.keystore = sqliteshelf.SQLiteShelf(filename, table="wallet")
         
         # Keep the blockchain
         self.blockchain = blockchain
+        
+        # We need a lock to protect our keystore from multithreaded access.
+        # TODO: Shrink the critical sections.
+        self.lock = threading.RLock()
     
     def generate_address(self):
         """
@@ -773,7 +799,6 @@ class Wallet(object):
         
         # Save it to the keystore
         self.keystore[hashlib.sha256(keypair.get_pubkey()).digest()] = keypair
-        self.keystore.sync()
         
     def get_address(self):
         """
@@ -781,12 +806,13 @@ class Wallet(object):
         
         """
         
-        if len(self.keystore) == 0:
-            # We need to make an address
-            self.generate_address()    
-        
-        # Just use the first address we have
-        return self.keystore.keys()[0]
+        with self.lock:
+            if len(self.keystore) == 0:
+                # We need to make an address
+                self.generate_address()    
+            
+            # Just use the first address we have
+            return self.keystore.keys()[0]
         
     def get_spendable_outputs(self):
         """
@@ -795,16 +821,18 @@ class Wallet(object):
         
         """
         
-        # Accumulate every unused output we have the key for
-        spendable_outputs = []
-        
-        for output in self.blockchain.state.unused_outputs:
-            if self.keystore.has_key(output[3]):
-                # We have a key for the output's pubkey hash. We can spend this.
-                spendable_outputs.append(output)
-                
-        # Return all the outputs we can spend.
-        return spendable_outputs
+        with self.lock:
+            # Accumulate every unused output we have the key for
+            spendable_outputs = []
+            
+            for output in self.blockchain.state.unused_outputs:
+                if self.keystore.has_key(output[3]):
+                    # We have a key for the output's pubkey hash. We can spend
+                    # this.
+                    spendable_outputs.append(output)
+                    
+            # Return all the outputs we can spend.
+            return spendable_outputs
     
     def get_balance(self):
         """
@@ -840,70 +868,72 @@ class Wallet(object):
         
         """
         
-        if not amount > 0:
-            # Transaction is unreasonable: not sending any coins anywhere.
-            return None
-            
-        # Make a transaction
-        transaction = Transaction()
+        with self.lock:
         
-        # This holds how much we have accumulated from the spendable outputs
-        # we've added to the transaction's inputs.
-        coins_collected = 0
-        
-        # This holds the set of public key hashes that we need to sign the
-        # transaction with.
-        key_hashes = set()
-        
-        for spendable in self.get_spendable_outputs():
-            # Unpack the amount we get from this as an input, and the key we
-            # need to use to spend it.
-            _, _, input_amount, key_needed = spendable
-        
-            # Add the unspent output as an input to the transaction
-            transaction.add_input(*spendable)
-            
-            # Say we've collected that many coins
-            coins_collected += input_amount
-            
-            # Say we need to sign with the appropriate key
-            key_hashes.add(key_needed)
-            
-            if coins_collected >= amount + fee:
-                # We have enough coins.
-                break
+            if not amount > 0:
+                # Transaction is unreasonable: not sending any coins anywhere.
+                return None
                 
-        if coins_collected < amount + fee:
-            # We couldn't find enough money for this transaction!
-            return None
+            # Make a transaction
+            transaction = Transaction()
             
-        # We've made a transaction with enough inputs!
-        # Add the outputs.
-        # First the amount we actually wanted to send.
-        transaction.add_output(amount, destination)
-        # Then the change back to us at some address we can receive on.
-        transaction.add_output(coins_collected - amount - fee, 
-            self.get_address())
-        # The fee should be left over.
-        
-        # Now do the authorizations. What do we need to sign?
-        to_sign = transaction.header_bytes()
-        
-        for key_hash in key_hashes:
-            # Load the keypair
-            keypair = self.keystore[key_hash]
+            # This holds how much we have accumulated from the spendable outputs
+            # we've added to the transaction's inputs.
+            coins_collected = 0
             
-            # Grab the public key
-            public_key = keypair.get_pubkey()
+            # This holds the set of public key hashes that we need to sign the
+            # transaction with.
+            key_hashes = set()
             
-            # Make the signature
-            signature = keypair.sign(to_sign)
+            for spendable in self.get_spendable_outputs():
+                # Unpack the amount we get from this as an input, and the key we
+                # need to use to spend it.
+                _, _, input_amount, key_needed = spendable
             
-            # Add the authorization to the transaction
-            transaction.add_authorization(public_key, signature)
+                # Add the unspent output as an input to the transaction
+                transaction.add_input(*spendable)
+                
+                # Say we've collected that many coins
+                coins_collected += input_amount
+                
+                # Say we need to sign with the appropriate key
+                key_hashes.add(key_needed)
+                
+                if coins_collected >= amount + fee:
+                    # We have enough coins.
+                    break
+                    
+            if coins_collected < amount + fee:
+                # We couldn't find enough money for this transaction!
+                return None
+                
+            # We've made a transaction with enough inputs!
+            # Add the outputs.
+            # First the amount we actually wanted to send.
+            transaction.add_output(amount, destination)
+            # Then the change back to us at some address we can receive on.
+            transaction.add_output(coins_collected - amount - fee, 
+                self.get_address())
+            # The fee should be left over.
             
-        # Nw the transaction is done!
-        return transaction
+            # Now do the authorizations. What do we need to sign?
+            to_sign = transaction.header_bytes()
+            
+            for key_hash in key_hashes:
+                # Load the keypair
+                keypair = self.keystore[key_hash]
+                
+                # Grab the public key
+                public_key = keypair.get_pubkey()
+                
+                # Make the signature
+                signature = keypair.sign(to_sign)
+                
+                # Add the authorization to the transaction
+                transaction.add_authorization(public_key, signature)
+                
+            # Nw the transaction is done!
+            return transaction
 
 if __name__ == "__main__":
     # Do a transaction test
